#include <cstdio>
#include <string>
#include <fstream>
#include <map>
#include <list>
#include <cassert>

#include "node.h"
#include "leafnode.h"

struct Context
{
	Context() : inputFileSize(0), root(nullptr)
	{
	}

	std::string inputFilePath;
	unsigned inputFileSize;
	std::map<char, unsigned> occurrences;
	Node *root;
	std::map<char, std::string> dict;
};

void countSymbols(Context &ctx)
{
	std::ifstream in(ctx.inputFilePath, std::ifstream::binary);

	if (!in.good())
	{
		fprintf(stderr, "Error opening \"%s\" for input\n", ctx.inputFilePath.c_str());
		return;
	}

	while (in.good())
	{
		char c;
		if (in.get(c))
		{
			ctx.inputFileSize++;
			ctx.occurrences[c]++;
		}
	}
}

void buildTree(Context &ctx)
{
	std::list<Node*> nodes;

	for (std::map<char, unsigned>::const_iterator it = ctx.occurrences.cbegin(); it != ctx.occurrences.cend(); it++)
	{
		LeafNode *n = new LeafNode(it->first, it->second);
		nodes.push_back(n);
	}
	nodes.sort(&Node::compare);

	while (nodes.size() > 1)
	{
		Node *left = nodes.back();
		nodes.pop_back();
		Node *right = nodes.back();
		nodes.pop_back();
		Node *n = new Node(left, right);

		nodes.push_back(n);
		nodes.sort(&Node::compare);
	}

	ctx.root = nodes.front();
	// Value of root node must match number of input symbols
	assert(ctx.root->getValue() == static_cast<unsigned>(ctx.inputFileSize));
}

void buildDictionary(Context &ctx)
{
	ctx.root->addCode(ctx.dict, "");
}

unsigned longestPath(const Node *node)
{
	unsigned left = node->getLeftChild() ? longestPath(node->getLeftChild()) : 0;
	unsigned right = node->getRightChild() ? longestPath(node->getRightChild()) : 0;
	return (left > right ? left : right) + 1;
}

unsigned smallestCodeLength(const Context &ctx)
{
	unsigned result = longestPath(ctx.root);

	for (std::map<char, std::string>::const_iterator it = ctx.dict.cbegin(); it != ctx.dict.cend(); it++)
	{
		if (it->second.size() < result)
			result = static_cast<unsigned>(it->second.size());
	}

	return result;
}

double averageCodeLength(const Context &ctx)
{
	unsigned sum = 0;

	for (std::map<char, std::string>::const_iterator it = ctx.dict.cbegin(); it != ctx.dict.cend(); it++)
	{
		sum += static_cast<unsigned>(it->second.size());
	}

	return sum / static_cast<double>(ctx.dict.size());
}

unsigned greatestCodeLength(const Context &ctx)
{
	unsigned result = 0;

	for (std::map<char, std::string>::const_iterator it = ctx.dict.cbegin(); it != ctx.dict.cend(); it++)
	{
		if (it->second.size() > result)
			result = static_cast<unsigned>(it->second.size());
	}

	return result;
}

unsigned encodedTextLength(const Context &ctx)
{
	unsigned result = 0;

	for (std::map<char, unsigned>::const_iterator it = ctx.occurrences.cbegin(); it != ctx.occurrences.cend(); it++)
	{
		std::map<char, std::string>::const_iterator code = ctx.dict.find(it->first);
		assert(code != ctx.dict.cend());
		result += static_cast<unsigned>(it->second * code->second.size());
	}

	return result;
}

void printStats(const Context &ctx)
{
	printf("Input file:           %s\n", ctx.inputFilePath.c_str());
	printf("Symbols (total):      %u\n", ctx.inputFileSize);
	printf("Symbols (individual): %u\n", static_cast<unsigned>(ctx.occurrences.size()));
	printf("Longest path:         %u\n", longestPath(ctx.root));
	printf("Code length\n");
	printf("    smallest:         %u\n", smallestCodeLength(ctx));
	printf("    average:          %.1f\n", averageCodeLength(ctx));
	printf("    greatest:         %u\n", greatestCodeLength(ctx));
	printf("Encoded text length\n");
	unsigned encodedLengthBits = encodedTextLength(ctx);
	unsigned encodedLengthBytes = (encodedLengthBits - 1) / 8 + 1;
	printf("    bits:             %u\n", encodedLengthBits);
	printf("    bytes:            %u\n", encodedLengthBytes);
	printf("Encoded/original:     %.1f%%\n", (encodedLengthBytes * 100.0) / ctx.inputFileSize);
}

void printTree(const Node *node, unsigned level = 0)
{
	if (node->getRightChild())
		printTree(node->getRightChild(), level + 1);
	node->print(level);
	if (node->getLeftChild())
		printTree(node->getLeftChild(), level + 1);
}

void printDictionary(const Context &ctx)
{
	printf("Dictionary:\n");
	for (std::map<char, std::string>::const_iterator it = ctx.dict.cbegin(); it != ctx.dict.cend(); it++)
	{
		uint8_t code = it->first;
		char c = isprint(code) ? code : '.';
		printf("    '%c' (0x%.2x): %s\n", c, code, it->second.c_str());
	}
}

int main(int argc, char *argv[])
{
	int result = EXIT_FAILURE;

	if (argc == 2)
	{
		Context ctx;

		ctx.inputFilePath = argv[1];
		countSymbols(ctx);
		if (ctx.occurrences.size() > 0)
		{
			buildTree(ctx);
			buildDictionary(ctx);
			printStats(ctx);

			//printTree(ctx.root);
			//printDictionary(ctx);
		}
		result = EXIT_SUCCESS;
	}
	else
	{
		fprintf(stderr, "Missing argument: file path\n");
	}

	return result;
}
